--[[
LibGuessRole-1.0 - Guess player roles.
Copyright (C) 2010 Adirelle

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
    * Redistribution of a stand alone version is strictly prohibited without
      prior written authorization from the LibGuessRole project manager.
    * Neither the name of the LibGuessRole authors nor the names of its contributors
      may be used to endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
--]]

local MAJOR, MINOR = "LibGuessRole-1.0", 8
assert(LibStub, MAJOR.." requires LibStub")
local lib = LibStub:NewLibrary(MAJOR, MINOR)
if not lib then return end

-- Source fiability constants, higher is better

local SOURCE_NONE = 0 -- Not
local SOURCE_CLASSGUESS = 10 -- Loose guess from class
local SOURCE_SPELLGUESS = 20 -- Guess from cast spells and self buffs
local SOURCE_INSPECT = 30 -- Guess from talent query
local SOURCE_PLAYER = 40 -- Guess from player talents
local SOURCE_IMMUTABLE = 50 -- Class-based role that cannot change

lib.SOURCE_NONE = SOURCE_NONE
lib.SOURCE_CLASSGUESS = SOURCE_CLASSGUESS
lib.SOURCE_SPELLGUESS = SOURCE_SPELLGUESS
lib.SOURCE_INSPECT = SOURCE_INSPECT
lib.SOURCE_PLAYER = SOURCE_PLAYER
lib.SOURCE_IMMUTABLE = SOURCE_IMMUTABLE

-- Role constants

local ROLE_MELEE = 'melee'
local ROLE_CASTER = 'caster'
local ROLE_TANK = 'tank'
local ROLE_HEALER = 'healer'

lib.ROLE_MELEE = ROLE_MELEE
lib.ROLE_CASTER = ROLE_CASTER
lib.ROLE_TANK = ROLE_TANK
lib.ROLE_HEALER = ROLE_HEALER

-- Data

if not lib.frame then
	lib.frame = CreateFrame('Frame')
	lib.frame:SetScript('OnEvent', function(self, event, ...)
		return lib[event](lib, ...)
	end)
	lib.frame:SetScript('OnUpdate', function(self, ...) lib:OnUpdate(...) end)
	lib.frame:Hide()
end

if not lib.callbacks then
	lib.callbacks = LibStub('CallbackHandler-1.0'):New(lib)
end

if not lib.NotifyInspectHook then
	hooksecurefunc('NotifyInspect', function(...) lib:NotifyInspectHook(...) end)
end

if not lib.roles then
	lib.roles = {}
end

if not lib.sources then
	lib.sources = setmetatable({}, {__index = function(t, k) return SOURCE_NONE end})
end

if not lib.timestamps then
	lib.timestamps = setmetatable({}, {__index = function(t, k) return 0 end})
end

if not lib.inspectQueue then
	lib.inspectQueue = {}
end

if not lib.nextInspectTime then
	lib.nextInspectTime = 0
end

lib.MAX_INSPECT_AGE = 60

-- Locals

local roles = lib.roles
local sources = lib.sources
local timestamps = lib.timestamps
local inspectQueue = lib.inspectQueue
local frame = lib.frame

-- Class-based guess

local immutableRoles = {
	MAGE = ROLE_CASTER,
	HUNTER = ROLE_MELEE,
	ROGUE = ROLE_MELEE,
	WARLOCK = ROLE_CASTER,
}

local classGuess = {
	DEATHKNIGHT = ROLE_MELEE,
	DRUID = ROLE_CASTER,
	PALADIN = function(unit)
		if UnitLevel(unit) == 80 and UnitManaMax(unit) > 15000 then
			return ROLE_HEALER
		elseif UnitAura(unit, (GetSpellInfo(25780))) then -- Righteous fury
			return ROLE_TANK
		else
			return ROLE_MELEE
		end
	end,
	PRIEST = ROLE_HEALER,
	SHAMAN = ROLE_MELEE,
	WARRIOR = function(unit)
		if UnitLevel(unit) == 80 and UnitHealthMax(unit) > 40000 then
			return ROLE_TANK
		else
			return ROLE_MELEE
		end
	end
}

function lib:GuessFromClass(unit)
	local _, class = UnitClass(unit)
	local role, source = immutableRoles[class], SOURCE_IMMUTABLE
	if not role then
		role, source = classGuess[class], SOURCE_CLASSGUESS
		if type(role) == "function" then
			role = role(unit)
		end
	end
	return role, source
end

-- Spell-based guess

local spellMap = lib.spellMap
if not spellMap then
	spellMap = {
		-- Paladin
		[20473] = ROLE_HEALER, -- Holy Shock
		[53563] = ROLE_HEALER, -- Beacon of Light
		[31842] = ROLE_HEALER, -- Divine Favor
		[20925] = ROLE_TANK,   -- Holy Shield
		[31935] = ROLE_TANK,   -- Avenger's Shield
		[53595] = ROLE_TANK,   -- Hammer of the Righteous
		[53385] = ROLE_MELEE,  -- Divine Storm
		[20066] = ROLE_MELEE,  -- Repentance

		-- Priest
		[47540] = ROLE_HEALER, -- Penance
		[34861] = ROLE_HEALER, -- Circle of Healing
		[33206] = ROLE_HEALER, -- Pain Suppression
		[15473] = ROLE_CASTER, -- Shadowform
		[34914] = ROLE_CASTER, -- Vampiric Touch
		[47585] = ROLE_CASTER, -- Dispersion
		[64044] = ROLE_CASTER, -- Psychic Horror
		[15487] = ROLE_CASTER, -- Silence

		-- Druid
		[48438] = ROLE_HEALER, -- Wild Growth
		[33891] = ROLE_HEALER, -- Tree of Life
		[18562] = ROLE_HEALER, -- Swift Mend
		[17116] = ROLE_HEALER, -- Nature's Swiftness
		[6807]  = ROLE_TANK,   -- Maul
		[779]   = ROLE_TANK,   -- Swipe (Bear Form)
		[80313] = ROLE_TANK,   -- Pulverize
		[5221]  = ROLE_MELEE,  -- Shred
		[62078] = ROLE_MELEE,  -- Swipe (Cat Form)
		[24858] = ROLE_CASTER, -- Moonkin Form
		[50516] = ROLE_CASTER, -- Typhoon
		[50288] = ROLE_CASTER, -- Starfall
		[33831] = ROLE_CASTER, -- Force of Nature
		[78674] = ROLE_CASTER, -- Starsurge

		-- Death Knight
		--[48263] = ROLE_TANK,  -- Frost presence
		--[48266] = ROLE_MELEE, -- Blood presence
		--[48265] = ROLE_MELEE, -- Unholy presence

		-- Warrior
		[71]    = ROLE_TANK,  -- Defensive Stance
		[6572]  = ROLE_TANK,  -- Revenge
		[12294] = ROLE_MELEE, -- Mortal Strike
		[23881] = ROLE_MELEE, -- Bloodthirst

		-- Shaman
		[61295] = ROLE_HEALER,  -- Riptide
		[974]   = ROLE_HEALER,  -- Earth Shield
		[16190] = ROLE_HEALER,  -- Mana Tide Totem
		[51490] = ROLE_CASTER,  -- Thunderstorm
	}
	lib.spellMap = spellMap
end

function lib:GuessFromSpellcast(spellName)
	local role = spellMap[spellName]
	if role then
		return role, SOURCE_SPELLGUESS
	end
end

function lib:GuessFromBuffs(unit)
	for index = 1, 2000 do
		local name, _, _, _, _, _, _, caster, _, _, spellId = UnitBuff(unit, index)
		if not name then return end
		local role = caster and UnitIsUnit(unit, caster) and (spellMap[name] or spellMap[spellId])
		if role then
			return role, SOURCE_SPELLGUESS
		end
	end
end

-- Talent-based guess

local function HasTalent(tabIndex, talentIndex, inspect, talentGroup)
	local _, _, _, _, rank, maxRank = GetTalentInfo(tabIndex, talentIndex, inspect, false, talentGroup)
	return rank and rank == maxRank
end

function lib:GuessFromTalents(unit)
	local inspecting = not UnitIsUnit(unit, "player")
	local _, class = UnitClass(unit)
	local source = inspecting and SOURCE_INSPECT or SOURCE_PLAYER

	-- Use the Cata API to get the primary tree
	local talentGroup = GetActiveTalentGroup(inspecting)
	local primaryTree = GetPrimaryTalentTree(inspecting, false, talentGroup)
	if not primaryTree then return end

	if class == "DRUID" then
		if primaryTree == 1 then
			return ROLE_CASTER, source
		elseif primaryTree == 2 then
			-- Consider feral as tank if she/he has Natural Reaction and Thick Hide
			return HasTalent(2, 18, inspecting) and HasTalent(2, 11, inspecting) and ROLE_TANK or ROLE_MELEE, source
		else
			return ROLE_HEALER, source
		end
	elseif class == "SHAMAN" and primaryTree == 1 then
		return ROLE_CASTER, source
	elseif class == "PRIEST" and primaryTree == 3 then
		return ROLE_CASTER, source
	end

	-- Get the tree roles
	local role1, role2 = GetTalentTreeRoles(primaryTree, inspecting, false)
	if (not role1 or role1 == "NONE") and (not role2 or role2 == "NONE") then return end

	if role1 == "HEALER" or role2 == "HEALER" then
		return ROLE_HEALER, source
	elseif role1 == "TANK" or role2 == "TANK" then
		return ROLE_TANK, source
	end

	return ROLE_MELEE, source
end

-- Helpers

function lib:GetUnitData(unit, noInspect)
	local guid = unit and UnitIsPlayer(unit) and UnitLevel(unit) >= 10 and UnitGUID(unit)
	if not guid then return end
	local role, source = roles[guid], sources[guid]
	if source == SOURCE_INSPECT and time() - timestamps[guid] > lib.MAX_INSPECT_AGE then
		source = SOURCE_NONE
	end
	if source == SOURCE_NONE then
		role, source = self:GuessFromClass(unit)
		if source < SOURCE_SPELLGUESS then
			local newRole, newSource = self:GuessFromBuffs(unit)
			if newRole then
				role, source = newRole, newSource
			end
		end
	end
	if not noInspect and source < SOURCE_INSPECT and not UnitIsUnit(unit, 'player') then
		if inspectQueue[guid] ~= unit then
			inspectQueue[guid] = unit
		end
		self:ResumeBackgroundQueries()
	end
	return guid, role, source
end

function lib:SetNewRole(guid, role, source)
	if guid and role and source then
		if source >= SOURCE_INSPECT and inspectQueue[guid] then
			inspectQueue[guid] = nil
		end
		if source >= sources[guid] then
			timestamps[guid] = time()
			sources[guid] = source
			if role ~= roles[guid] then
				roles[guid] = role
				if self.initialized then
					self.callbacks:Fire('LibGuessRole_RoleChanged', guid, role, source)
				end
			end
		end
	end
end

-- Catch talent queries

function lib:NotifyInspectHook(unit)
	if CanInspect(unit) then
		self.nextInspectTime = time() + 2
		self.callbacks:Fire('LibGuessRole_NotifyInspect', unit)
	end
end

local function GetUnitForGUID(guid)
	if guid == UnitGUID("target") then
		return "target"
	elseif guid == UnitGUID("mouseover") then
		return "mouseover"
	elseif guid == UnitGUID("focus") then
		return "focus"
	end
	for i = 1, GetNumPartyMembers() do
		local unit = "party"..i
		if UnitGUID(unit) == guid then
			return unit
		end
	end
	for i = 1, GetNumRaidMembers() do
		local unit = "raid"..i
		if UnitGUID(unit) == guid then
			return unit
		end
	end
end

function lib:INSPECT_READY(guid)
	local unit = GetUnitForGUID(guid)
	if unit then
		local guid, _, source = self:GetUnitData(unit, true)
		if guid and source <= SOURCE_INSPECT then
			self:SetNewRole(guid, self:GuessFromTalents(unit))
		end
		self.callbacks:Fire('LibGuessRole_TalentsAvailable', unit)
	end
end

-- Handle auras and buffs

function lib:UNIT_SPELLCAST_SENT(_, unit, spell)
	local guid, _, source = self:GetUnitData(unit)
	if guid and source <= SOURCE_SPELLGUESS then
		self:SetNewRole(guid, self:GuessFromSpellcast(spell))
	end
end

function lib:UNIT_AURA(_, unit)
	local guid, _, source = self:GetUnitData(unit)
	if guid and source <= SOURCE_SPELLGUESS then
		self:SetNewRole(guid, self:GuessFromBuffs(unit))
	end
end

-- Handle local player

function lib:PLAYER_TALENT_UPDATE()
	local guid, role, source = self:GetUnitData("player")
	if guid then -- UnitGUID("player") can be nil when (s)he is "out of the world"
		if source <= SOURCE_PLAYER then
			self:SetNewRole(guid, self:GuessFromTalents("player"))
		elseif source == SOURCE_IMMUTABLE then
			frame:UnregisterEvent('ACTIVE_TALENT_GROUP_CHANGED')
			frame:UnregisterEvent('PLAYER_TALENT_UPDATE')
			lib.PLAYER_TALENT_UPDATE = nil
			lib.ACTIVE_TALENT_GROUP_CHANGED = nil
		end
	end
end
lib.ACTIVE_TALENT_GROUP_CHANGED = lib.PLAYER_TALENT_UPDATE

-- Handle inspect queue

function lib:OnUpdate()
	if self:ResumeBackgroundQueries() and self.nextInspectTime < time() then
		for guid, unit in pairs(inspectQueue) do
			if UnitGUID(unit) ~= guid then
				inspectQueue[guid] = nil
			elseif CanInspect(unit) then
				NotifyInspect(unit)
				return
			end
		end
		-- Could not inspect anyone, try again in one second
		self.nextInspectTime = time() + 1
	end
end

function lib:PauseBackgroundQueries()
	frame:Hide()
end

function lib:ResumeBackgroundQueries()
	if not InCombatLockdown() and next(inspectQueue) and not (InspectFrame and InspectFrame:IsShown()) then
		frame:Show()
		return true
	else
		lib:PauseBackgroundQueries()
	end
end

lib.PLAYER_REGEN_DISABLED = lib.PauseBackgroundQueries
lib.PLAYER_LEAVING_WORLD = lib.PauseBackgroundQueries

lib.PLAYER_REGEN_ENABLED = lib.ResumeBackgroundQueries
lib.PLAYER_ENTERING_WORLD = lib.ResumeBackgroundQueries

-- API

function lib:GetUnitRole(unit)
	local guid, role, source = self:GetUnitData(unit)
	if guid and source > SOURCE_NONE then
		if not roles[guid] then
			self:SetNewRole(guid, role, source)
		end
		return role, source
	end
end

if not lib.initialized then
	-- Library initialization

	function lib:PLAYER_LOGIN()
		frame:UnregisterEvent('PLAYER_LOGIN')
		self.PLAYER_LOGIN = nil

		-- Get spell names
		for id, role in pairs(spellMap) do
			if type(id) == "number" then
				local name = GetSpellInfo(id)
				if name then
					spellMap[name] = role
				else
					geterrorhandler()(format("Unknown spell #%d for %s", id, role))
				end
			end
		end

		-- Register events
		frame:RegisterEvent('PLAYER_REGEN_DISABLED')
		frame:RegisterEvent('PLAYER_REGEN_ENABLED')
		frame:RegisterEvent('PLAYER_ENTERING_WORLD')
		frame:RegisterEvent('PLAYER_LEAVING_WORLD')

		frame:RegisterEvent('ACTIVE_TALENT_GROUP_CHANGED')
		frame:RegisterEvent('PLAYER_TALENT_UPDATE')

		frame:RegisterEvent('UNIT_SPELLCAST_SENT')
		frame:RegisterEvent('UNIT_AURA')

		frame:RegisterEvent('INSPECT_READY')

		-- Attempt to force player role detection from talents
		self:PLAYER_TALENT_UPDATE()

		if self.PLAYER_TALENT_UPDATE then
			-- Player role is dependant on talents but they are not available until first PLAYER_ALIVE
			frame:RegisterEvent('PLAYER_ALIVE')
			function self:PLAYER_ALIVE()
				frame:UnregisterEvent('PLAYER_ALIVE')
				self.PLAYER_ALIVE = nil
				return self:PLAYER_TALENT_UPDATE()
			end
		end

		self.initialized = true
	end

	if IsLoggedIn() then
		lib:PLAYER_LOGIN()
	else
		frame:RegisterEvent('PLAYER_LOGIN')
	end
end
